@@ -95,12 +95,7 @@ int Http2DownstreamConnection::attach_downstream(Downstream *downstream) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
   http2session_->add_downstream_connection(this);
-  if (http2session_->get_state() == Http2Session::DISCONNECTED) {
-    http2session_->signal_write();
-    if (http2session_->get_state() == Http2Session::DISCONNECTED) {
-      return -1;
-    }
-  }
+  http2session_->signal_write();
 
   downstream_ = downstream;
   downstream_->reset_downstream_rtimer();
@@ -108,7 +103,9 @@ int Http2DownstreamConnection::attach_downstream(Downstream *downstream) {
   auto &req = downstream_->request();
 
   // HTTP/2 disables HTTP Upgrade.
-  req.upgrade_request = false;
+  if (req.method != HTTP_CONNECT) {
+    req.upgrade_request = false;
+  }
 
   return 0;
 }
@@ -120,11 +117,11 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream) {
 
   auto &resp = downstream_->response();
 
-  if (submit_rst_stream(downstream) == 0) {
-    http2session_->signal_write();
-  }
-
   if (downstream_->get_downstream_stream_id() != -1) {
+    if (submit_rst_stream(downstream) == 0) {
+      http2session_->signal_write();
+    }
+
     http2session_->consume(downstream_->get_downstream_stream_id(),
                            resp.unconsumed_body_length);
 
@@ -182,24 +179,11 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
   }
   const auto &req = downstream->request();
   auto input = downstream->get_request_buf();
-  auto nread = input->remove(buf, length);
-  auto input_empty = input->rleft() == 0;
 
-  if (nread > 0) {
-    // This is important because it will handle flow control
-    // stuff.
-    if (downstream->get_upstream()->resume_read(SHRPX_NO_BUFFER, downstream,
-                                                nread) != 0) {
-      // In this case, downstream may be deleted.
-      return NGHTTP2_ERR_CALLBACK_FAILURE;
-    }
+  auto nread = std::min(input->rleft(), length);
+  auto input_empty = input->rleft() == nread;
 
-    // Check dconn is still alive because Upstream::resume_read()
-    // may delete downstream which will delete dconn.
-    if (sd->dconn == nullptr) {
-      return NGHTTP2_ERR_DEFERRED;
-    }
-  }
+  *data_flags |= NGHTTP2_DATA_FLAG_NO_COPY;
 
   if (input_empty &&
       downstream->get_request_state() == Downstream::MSG_COMPLETE &&
@@ -232,12 +216,6 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
     }
   }
 
-  if (!input_empty) {
-    downstream->reset_downstream_wtimer();
-  } else {
-    downstream->disable_downstream_wtimer();
-  }
-
   if (nread == 0 && (*data_flags & NGHTTP2_DATA_FLAG_EOF) == 0) {
     downstream->disable_downstream_wtimer();
 
@@ -268,11 +246,11 @@ int Http2DownstreamConnection::push_request_headers() {
 
   auto &balloc = downstream_->get_block_allocator();
 
-  auto &httpconf = get_config()->http;
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &httpconf = config->http;
+  auto &http2conf = config->http2;
 
-  auto no_host_rewrite = httpconf.no_host_rewrite ||
-                         get_config()->http2_proxy ||
+  auto no_host_rewrite = httpconf.no_host_rewrite || config->http2_proxy ||
                          req.method == HTTP_CONNECT;
 
   // http2session_ has already in CONNECTED state, so we can get
@@ -333,13 +311,6 @@ int Http2DownstreamConnection::push_request_headers() {
 
   http2::copy_headers_to_nva_nocopy(nva, req.fs.headers());
 
-  bool chunked_encoding = false;
-  auto transfer_encoding = req.fs.header(http2::HD_TRANSFER_ENCODING);
-  if (transfer_encoding &&
-      util::strieq_l("chunked", (*transfer_encoding).value)) {
-    chunked_encoding = true;
-  }
-
   if (!http2conf.no_cookie_crumbling) {
     downstream_->crumble_request_cookie(nva);
   }
@@ -355,7 +326,7 @@ int Http2DownstreamConnection::push_request_headers() {
   if (fwdconf.params) {
     auto params = fwdconf.params;
 
-    if (get_config()->http2_proxy || req.method == HTTP_CONNECT) {
+    if (config->http2_proxy || req.method == HTTP_CONNECT) {
       params &= ~FORWARDED_PROTO;
     }
 
@@ -386,7 +357,7 @@ int Http2DownstreamConnection::push_request_headers() {
 
   if (xffconf.add) {
     StringRef xff_value;
-    auto addr = StringRef{upstream->get_client_handler()->get_ipaddr()};
+    const auto &addr = upstream->get_client_handler()->get_ipaddr();
     if (xff) {
       xff_value = concat_string_ref(balloc, xff->value,
                                     StringRef::from_lit(", "), addr);
@@ -398,7 +369,7 @@ int Http2DownstreamConnection::push_request_headers() {
     nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-for", xff->value));
   }
 
-  if (!get_config()->http2_proxy && req.method != HTTP_CONNECT) {
+  if (!config->http2_proxy && req.method != HTTP_CONNECT) {
     // We use same protocol with :scheme header field
     nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", req.scheme));
   }
@@ -442,16 +413,18 @@ int Http2DownstreamConnection::push_request_headers() {
   if (LOG_ENABLED(INFO)) {
     std::stringstream ss;
     for (auto &nv : nva) {
-      ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
+      ss << TTY_HTTP_HD << StringRef{nv.name, nv.namelen} << TTY_RST << ": "
+         << StringRef{nv.value, nv.valuelen} << "\n";
     }
     DCLOG(INFO, this) << "HTTP request headers\n" << ss.str();
   }
 
-  auto content_length = req.fs.header(http2::HD_CONTENT_LENGTH);
-  // TODO check content-length: 0 case
+  auto transfer_encoding = req.fs.header(http2::HD_TRANSFER_ENCODING);
 
-  if (req.method == HTTP_CONNECT || chunked_encoding || content_length ||
-      req.http2_expect_body) {
+  // Add body as long as transfer-encoding is given even if
+  // req.fs.content_length == 0 to forward trailer fields.
+  if (req.method == HTTP_CONNECT || transfer_encoding ||
+      req.fs.content_length > 0 || req.http2_expect_body) {
     // Request-body is expected.
     nghttp2_data_provider data_prd{{}, http2_data_read_callback};
     rv = http2session_->submit_request(this, nva.data(), nva.size(), &data_prd);
@@ -506,8 +479,7 @@ int Http2DownstreamConnection::resume_read(IOCtrlReason reason,
                                            size_t consumed) {
   int rv;
 
-  if (http2session_->get_state() != Http2Session::CONNECTED ||
-      !http2session_->get_flow_control()) {
+  if (http2session_->get_state() != Http2Session::CONNECTED) {
     return 0;
   }
 
@@ -566,9 +538,11 @@ int Http2DownstreamConnection::on_timeout() {
   return submit_rst_stream(downstream_, NGHTTP2_NO_ERROR);
 }
 
-DownstreamAddrGroup *
+const std::shared_ptr<DownstreamAddrGroup> &
 Http2DownstreamConnection::get_downstream_addr_group() const {
   return http2session_->get_downstream_addr_group();
 }
 
+DownstreamAddr *Http2DownstreamConnection::get_addr() const { return nullptr; }
+
 } // namespace shrpx
