@@ -90,7 +90,10 @@ void connchk_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 namespace {
 void settings_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto http2session = static_cast<Http2Session *>(w->data);
-  SSLOG(INFO, http2session) << "SETTINGS timeout";
+
+  if (LOG_ENABLED(INFO)) {
+    SSLOG(INFO, http2session) << "SETTINGS timeout";
+  }
 
   downstream_failure(http2session->get_addr());
 
@@ -261,25 +264,28 @@ int Http2Session::disconnect(bool hard) {
   state_ = DISCONNECTED;
 
   // When deleting Http2DownstreamConnection, it calls this object's
-  // remove_downstream_connection(). The multiple
+  // remove_downstream_connection().  The multiple
   // Http2DownstreamConnection objects belong to the same
-  // ClientHandler object.  So first dump ClientHandler objects.
+  // ClientHandler object if upstream is h2 or SPDY.  So be careful
+  // when you delete ClientHandler here.
   //
   // We allow creating new pending Http2DownstreamConnection with this
   // object.  Upstream::on_downstream_reset() may add
-  // Http2DownstreamConnection.
+  // Http2DownstreamConnection to another Http2Session.
 
-  std::set<ClientHandler *> handlers;
-  for (auto dc = dconns_.head; dc; dc = dc->dlnext) {
-    if (!dc->get_client_handler()) {
-      continue;
-    }
-    handlers.insert(dc->get_client_handler());
-  }
-  for (auto h : handlers) {
-    if (h->get_upstream()->on_downstream_reset(hard) != 0) {
-      delete h;
+  for (auto dc = dconns_.head; dc;) {
+    auto next = dc->dlnext;
+    auto downstream = dc->get_downstream();
+    auto upstream = downstream->get_upstream();
+
+    // Failure is allowed only for HTTP/1 upstream where upstream is
+    // not shared by multiple Downstreams.
+    if (upstream->on_downstream_reset(downstream, hard) != 0) {
+      delete upstream->get_client_handler();
     }
+
+    // dc was deleted
+    dc = next;
   }
 
   auto streams = std::move(streams_);
@@ -307,6 +313,8 @@ int Http2Session::initiate_connection() {
     }
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
   const auto &proxy = get_config()->downstream_http_proxy;
   if (!proxy.host.empty() && state_ == DISCONNECTED) {
     if (LOG_ENABLED(INFO)) {
@@ -345,7 +353,7 @@ int Http2Session::initiate_connection() {
 
     conn_.wlimit.startw();
 
-    // TODO we should have timeout for connection establishment
+    conn_.wt.repeat = downstreamconf.timeout.connect;
     ev_timer_again(conn_.loop, &conn_.wt);
 
     write_ = &Http2Session::connected;
@@ -478,7 +486,8 @@ int Http2Session::initiate_connection() {
     if (state_ != CONNECTED) {
       state_ = CONNECTING;
       conn_.wlimit.startw();
-      // TODO we should have timeout for connection establishment
+
+      conn_.wt.repeat = downstreamconf.timeout.connect;
       ev_timer_again(conn_.loop, &conn_.wt);
     } else {
       conn_.rlimit.startw();
@@ -731,46 +740,44 @@ int on_stream_close_callback(nghttp2_session *session, int32_t stream_id,
   auto dconn = sd->dconn;
   if (dconn) {
     auto downstream = dconn->get_downstream();
-    if (downstream && downstream->get_downstream_stream_id() == stream_id) {
-      auto upstream = downstream->get_upstream();
+    auto upstream = downstream->get_upstream();
 
-      if (downstream->get_downstream_stream_id() % 2 == 0 &&
-          downstream->get_request_state() == Downstream::INITIAL) {
-        // Downstream is canceled in backend before it is submitted in
-        // frontend session.
+    if (downstream->get_downstream_stream_id() % 2 == 0 &&
+        downstream->get_request_state() == Downstream::INITIAL) {
+      // Downstream is canceled in backend before it is submitted in
+      // frontend session.
 
-        // This will avoid to send RST_STREAM to backend
-        downstream->set_response_state(Downstream::MSG_RESET);
-        upstream->cancel_premature_downstream(downstream);
-      } else {
-        if (downstream->get_upgraded() &&
-            downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
-          // For tunneled connection, we have to submit RST_STREAM to
-          // upstream *after* whole response body is sent. We just set
-          // MSG_COMPLETE here. Upstream will take care of that.
-          downstream->get_upstream()->on_downstream_body_complete(downstream);
-          downstream->set_response_state(Downstream::MSG_COMPLETE);
-        } else if (error_code == NGHTTP2_NO_ERROR) {
-          switch (downstream->get_response_state()) {
-          case Downstream::MSG_COMPLETE:
-          case Downstream::MSG_BAD_HEADER:
-            break;
-          default:
-            downstream->set_response_state(Downstream::MSG_RESET);
-          }
-        } else if (downstream->get_response_state() !=
-                   Downstream::MSG_BAD_HEADER) {
+      // This will avoid to send RST_STREAM to backend
+      downstream->set_response_state(Downstream::MSG_RESET);
+      upstream->cancel_premature_downstream(downstream);
+    } else {
+      if (downstream->get_upgraded() &&
+          downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
+        // For tunneled connection, we have to submit RST_STREAM to
+        // upstream *after* whole response body is sent. We just set
+        // MSG_COMPLETE here. Upstream will take care of that.
+        downstream->get_upstream()->on_downstream_body_complete(downstream);
+        downstream->set_response_state(Downstream::MSG_COMPLETE);
+      } else if (error_code == NGHTTP2_NO_ERROR) {
+        switch (downstream->get_response_state()) {
+        case Downstream::MSG_COMPLETE:
+        case Downstream::MSG_BAD_HEADER:
+          break;
+        default:
           downstream->set_response_state(Downstream::MSG_RESET);
         }
-        if (downstream->get_response_state() == Downstream::MSG_RESET &&
-            downstream->get_response_rst_stream_error_code() ==
-                NGHTTP2_NO_ERROR) {
-          downstream->set_response_rst_stream_error_code(error_code);
-        }
-        call_downstream_readcb(http2session, downstream);
+      } else if (downstream->get_response_state() !=
+                 Downstream::MSG_BAD_HEADER) {
+        downstream->set_response_state(Downstream::MSG_RESET);
       }
-      // dconn may be deleted
+      if (downstream->get_response_state() == Downstream::MSG_RESET &&
+          downstream->get_response_rst_stream_error_code() ==
+              NGHTTP2_NO_ERROR) {
+        downstream->set_response_rst_stream_error_code(error_code);
+      }
+      call_downstream_readcb(http2session, downstream);
     }
+    // dconn may be deleted
   }
   // The life time of StreamData ends here
   http2session->remove_stream_data(sd);
@@ -800,9 +807,6 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
     return 0;
   }
   auto downstream = sd->dconn->get_downstream();
-  if (!downstream) {
-    return 0;
-  }
 
   auto namebuf = nghttp2_rcbuf_get_buf(name);
   auto valuebuf = nghttp2_rcbuf_get_buf(value);
@@ -859,7 +863,7 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
         nghttp2_session_get_stream_user_data(session, promised_stream_id));
     if (!promised_sd || !promised_sd->dconn) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
-      return 0;
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     auto promised_downstream = promised_sd->dconn->get_downstream();
@@ -901,6 +905,43 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
 }
 } // namespace
 
+namespace {
+int on_invalid_header_callback2(nghttp2_session *session,
+                                const nghttp2_frame *frame, nghttp2_rcbuf *name,
+                                nghttp2_rcbuf *value, uint8_t flags,
+                                void *user_data) {
+  auto http2session = static_cast<Http2Session *>(user_data);
+  auto sd = static_cast<StreamData *>(
+      nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
+  if (!sd || !sd->dconn) {
+    return 0;
+  }
+
+  int32_t stream_id;
+
+  if (frame->hd.type == NGHTTP2_PUSH_PROMISE) {
+    stream_id = frame->push_promise.promised_stream_id;
+  } else {
+    stream_id = frame->hd.stream_id;
+  }
+
+  if (LOG_ENABLED(INFO)) {
+    auto namebuf = nghttp2_rcbuf_get_buf(name);
+    auto valuebuf = nghttp2_rcbuf_get_buf(value);
+
+    SSLOG(INFO, http2session)
+        << "Invalid header field for stream_id=" << stream_id
+        << " in frame type=" << static_cast<uint32_t>(frame->hd.type)
+        << ": name=[" << StringRef{namebuf.base, namebuf.len} << "], value=["
+        << StringRef{valuebuf.base, valuebuf.len} << "]";
+  }
+
+  http2session->submit_rst_stream(stream_id, NGHTTP2_PROTOCOL_ERROR);
+
+  return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+}
+} // namespace
+
 namespace {
 int on_begin_headers_callback(nghttp2_session *session,
                               const nghttp2_frame *frame, void *user_data) {
@@ -919,13 +960,6 @@ int on_begin_headers_callback(nghttp2_session *session,
                                       NGHTTP2_INTERNAL_ERROR);
       return 0;
     }
-    auto downstream = sd->dconn->get_downstream();
-    if (!downstream ||
-        downstream->get_downstream_stream_id() != frame->hd.stream_id) {
-      http2session->submit_rst_stream(frame->hd.stream_id,
-                                      NGHTTP2_INTERNAL_ERROR);
-      return 0;
-    }
     return 0;
   }
   case NGHTTP2_PUSH_PROMISE: {
@@ -934,7 +968,7 @@ int on_begin_headers_callback(nghttp2_session *session,
         nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
     if (!sd || !sd->dconn) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
-      return 0;
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     auto downstream = sd->dconn->get_downstream();
@@ -945,6 +979,7 @@ int on_begin_headers_callback(nghttp2_session *session,
     if (http2session->handle_downstream_push_promise(downstream,
                                                      promised_stream_id) != 0) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     return 0;
@@ -976,6 +1011,10 @@ int on_response_headers(Http2Session *http2session, Downstream *downstream,
   resp.http_major = 2;
   resp.http_minor = 0;
 
+  downstream->set_downstream_addr_group(
+      http2session->get_downstream_addr_group());
+  downstream->set_addr(http2session->get_addr());
+
   if (LOG_ENABLED(INFO)) {
     std::stringstream ss;
     for (auto &nv : nva) {
@@ -1019,8 +1058,8 @@ int on_response_headers(Http2Session *http2session, Downstream *downstream,
     }
     downstream->set_request_state(Downstream::HEADER_COMPLETE);
     if (LOG_ENABLED(INFO)) {
-      SSLOG(INFO, http2session)
-          << "HTTP upgrade success. stream_id=" << frame->hd.stream_id;
+      SSLOG(INFO, http2session) << "HTTP upgrade success. stream_id="
+                                << frame->hd.stream_id;
     }
   } else {
     auto content_length = resp.fs.header(http2::HD_CONTENT_LENGTH);
@@ -1082,11 +1121,6 @@ int on_frame_recv_callback(nghttp2_session *session, const nghttp2_frame *frame,
       return 0;
     }
     auto downstream = sd->dconn->get_downstream();
-    if (!downstream ||
-        downstream->get_downstream_stream_id() != frame->hd.stream_id) {
-      return 0;
-    }
-
     auto upstream = downstream->get_upstream();
     rv = upstream->on_downstream_body(downstream, nullptr, 0, true);
     if (rv != 0) {
@@ -1121,10 +1155,6 @@ int on_frame_recv_callback(nghttp2_session *session, const nghttp2_frame *frame,
     }
     auto downstream = sd->dconn->get_downstream();
 
-    if (!downstream) {
-      return 0;
-    }
-
     if (frame->headers.cat == NGHTTP2_HCAT_RESPONSE ||
         frame->headers.cat == NGHTTP2_HCAT_PUSH_RESPONSE) {
       rv = on_response_headers(http2session, downstream, session, frame);
@@ -1170,13 +1200,9 @@ int on_frame_recv_callback(nghttp2_session *session, const nghttp2_frame *frame,
         nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
     if (sd && sd->dconn) {
       auto downstream = sd->dconn->get_downstream();
-      if (downstream &&
-          downstream->get_downstream_stream_id() == frame->hd.stream_id) {
-
-        downstream->set_response_rst_stream_error_code(
-            frame->rst_stream.error_code);
-        call_downstream_readcb(http2session, downstream);
-      }
+      downstream->set_response_rst_stream_error_code(
+          frame->rst_stream.error_code);
+      call_downstream_readcb(http2session, downstream);
     }
     return 0;
   }
@@ -1278,9 +1304,7 @@ int on_data_chunk_recv_callback(nghttp2_session *session, uint8_t flags,
     return 0;
   }
   auto downstream = sd->dconn->get_downstream();
-  if (!downstream || downstream->get_downstream_stream_id() != stream_id ||
-      !downstream->expect_response_body()) {
-
+  if (!downstream->expect_response_body()) {
     http2session->submit_rst_stream(stream_id, NGHTTP2_INTERNAL_ERROR);
 
     if (http2session->consume(stream_id, len) != 0) {
@@ -1332,10 +1356,6 @@ int on_frame_send_callback(nghttp2_session *session, const nghttp2_frame *frame,
   auto http2session = static_cast<Http2Session *>(user_data);
 
   if (frame->hd.type == NGHTTP2_DATA || frame->hd.type == NGHTTP2_HEADERS) {
-    if ((frame->hd.flags & NGHTTP2_FLAG_END_STREAM) == 0) {
-      return 0;
-    }
-
     auto sd = static_cast<StreamData *>(
         nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
 
@@ -1345,8 +1365,12 @@ int on_frame_send_callback(nghttp2_session *session, const nghttp2_frame *frame,
 
     auto downstream = sd->dconn->get_downstream();
 
-    if (!downstream ||
-        downstream->get_downstream_stream_id() != frame->hd.stream_id) {
+    if (frame->hd.type == NGHTTP2_HEADERS &&
+        frame->headers.cat == NGHTTP2_HCAT_REQUEST) {
+      downstream->set_request_header_sent(true);
+    }
+
+    if ((frame->hd.flags & NGHTTP2_FLAG_END_STREAM) == 0) {
       return 0;
     }
 
@@ -1371,29 +1395,42 @@ int on_frame_not_send_callback(nghttp2_session *session,
   if (LOG_ENABLED(INFO)) {
     SSLOG(INFO, http2session) << "Failed to send control frame type="
                               << static_cast<uint32_t>(frame->hd.type)
-                              << "lib_error_code=" << lib_error_code << ":"
+                              << ", lib_error_code=" << lib_error_code << ": "
                               << nghttp2_strerror(lib_error_code);
   }
-  if (frame->hd.type == NGHTTP2_HEADERS &&
-      lib_error_code != NGHTTP2_ERR_STREAM_CLOSED &&
-      lib_error_code != NGHTTP2_ERR_STREAM_CLOSING) {
-    // To avoid stream hanging around, flag Downstream::MSG_RESET.
-    auto sd = static_cast<StreamData *>(
-        nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
-    if (!sd) {
-      return 0;
-    }
-    if (!sd->dconn) {
-      return 0;
-    }
-    auto downstream = sd->dconn->get_downstream();
-    if (!downstream ||
-        downstream->get_downstream_stream_id() != frame->hd.stream_id) {
-      return 0;
-    }
-    downstream->set_response_state(Downstream::MSG_RESET);
-    call_downstream_readcb(http2session, downstream);
+  if (frame->hd.type != NGHTTP2_HEADERS ||
+      lib_error_code == NGHTTP2_ERR_STREAM_CLOSED ||
+      lib_error_code == NGHTTP2_ERR_STREAM_CLOSING) {
+    return 0;
   }
+
+  auto sd = static_cast<StreamData *>(
+      nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
+  if (!sd) {
+    return 0;
+  }
+  if (!sd->dconn) {
+    return 0;
+  }
+  auto downstream = sd->dconn->get_downstream();
+
+  if (lib_error_code == NGHTTP2_ERR_START_STREAM_NOT_ALLOWED) {
+    // Migrate to another downstream connection.
+    auto upstream = downstream->get_upstream();
+
+    if (upstream->on_downstream_reset(downstream, false)) {
+      // This should be done for h1 upstream only.  Deleting
+      // ClientHandler for h2 or SPDY upstream may lead to crash.
+      delete upstream->get_client_handler();
+    }
+
+    return 0;
+  }
+
+  // To avoid stream hanging around, flag Downstream::MSG_RESET.
+  downstream->set_response_state(Downstream::MSG_RESET);
+  call_downstream_readcb(http2session, downstream);
+
   return 0;
 }
 } // namespace
@@ -1479,6 +1516,9 @@ nghttp2_session_callbacks *create_http2_downstream_callbacks() {
   nghttp2_session_callbacks_set_on_header_callback2(callbacks,
                                                     on_header_callback2);
 
+  nghttp2_session_callbacks_set_on_invalid_header_callback2(
+      callbacks, on_invalid_header_callback2);
+
   nghttp2_session_callbacks_set_on_begin_headers_callback(
       callbacks, on_begin_headers_callback);
 
@@ -1524,7 +1564,8 @@ int Http2Session::connection_made() {
     }
   }
 
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &http2conf = config->http2;
 
   rv = nghttp2_session_client_new2(&session_, http2conf.downstream.callbacks,
                                    this, http2conf.downstream.option);
@@ -1533,34 +1574,38 @@ int Http2Session::connection_made() {
     return -1;
   }
 
-  std::array<nghttp2_settings_entry, 3> entry;
+  std::array<nghttp2_settings_entry, 4> entry;
   size_t nentry = 2;
   entry[0].settings_id = NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS;
   entry[0].value = http2conf.downstream.max_concurrent_streams;
 
   entry[1].settings_id = NGHTTP2_SETTINGS_INITIAL_WINDOW_SIZE;
-  entry[1].value = (1 << http2conf.downstream.window_bits) - 1;
+  entry[1].value = http2conf.downstream.window_size;
 
-  if (http2conf.no_server_push || get_config()->http2_proxy) {
+  if (http2conf.no_server_push || config->http2_proxy) {
     entry[nentry].settings_id = NGHTTP2_SETTINGS_ENABLE_PUSH;
     entry[nentry].value = 0;
     ++nentry;
   }
 
+  if (http2conf.downstream.decoder_dynamic_table_size !=
+      NGHTTP2_DEFAULT_HEADER_TABLE_SIZE) {
+    entry[nentry].settings_id = NGHTTP2_SETTINGS_HEADER_TABLE_SIZE;
+    entry[nentry].value = http2conf.downstream.decoder_dynamic_table_size;
+    ++nentry;
+  }
+
   rv = nghttp2_submit_settings(session_, NGHTTP2_FLAG_NONE, entry.data(),
                                nentry);
   if (rv != 0) {
     return -1;
   }
 
-  if (http2conf.downstream.connection_window_bits != 16) {
-    int32_t window_size =
-        (1 << http2conf.downstream.connection_window_bits) - 1;
-    rv = nghttp2_session_set_local_window_size(session_, NGHTTP2_FLAG_NONE, 0,
-                                               window_size);
-    if (rv != 0) {
-      return -1;
-    }
+  rv = nghttp2_session_set_local_window_size(
+      session_, NGHTTP2_FLAG_NONE, 0,
+      http2conf.downstream.connection_window_size);
+  if (rv != 0) {
+    return -1;
   }
 
   auto must_terminate =
@@ -1764,7 +1809,8 @@ void Http2Session::submit_pending_requests() {
   for (auto dconn = dconns_.head; dconn; dconn = dconn->dlnext) {
     auto downstream = dconn->get_downstream();
 
-    if (!downstream || !downstream->request_submission_ready()) {
+    if (!downstream->get_request_pending() ||
+        !downstream->request_submission_ready()) {
       continue;
     }
 
@@ -1799,11 +1845,11 @@ int Http2Session::read_noop(const uint8_t *data, size_t datalen) { return 0; }
 int Http2Session::write_noop() { return 0; }
 
 int Http2Session::connected() {
-  if (!util::check_socket_connected(conn_.fd)) {
-    if (LOG_ENABLED(INFO)) {
-      SSLOG(INFO, this) << "Backend connect failed; addr="
-                        << util::to_numeric_addr(&addr_->addr);
-    }
+  auto sock_error = util::get_socket_error(conn_.fd);
+  if (sock_error != 0) {
+    SSLOG(WARN, this) << "Backend connect failed; addr="
+                      << util::to_numeric_addr(&addr_->addr)
+                      << ": errno=" << sock_error;
 
     downstream_failure(addr_);
 
@@ -1814,6 +1860,12 @@ int Http2Session::connected() {
     SSLOG(INFO, this) << "Connection established";
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  // Reset timeout for write.  Previously, we set timeout for connect.
+  conn_.wt.repeat = downstreamconf.timeout.write;
+  ev_timer_again(conn_.loop, &conn_.wt);
+
   conn_.rlimit.startw();
 
   read_ = &Http2Session::read_clear;
@@ -2102,6 +2154,7 @@ int Http2Session::handle_downstream_push_promise_complete(
   auto upstream = promised_downstream->get_upstream();
 
   promised_downstream->set_request_state(Downstream::MSG_COMPLETE);
+  promised_downstream->set_request_header_sent(true);
 
   if (upstream->on_downstream_push_promise_complete(downstream,
                                                     promised_downstream) != 0) {
@@ -2126,8 +2179,9 @@ bool Http2Session::max_concurrency_reached(size_t extra) const {
                  session_, NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS);
 }
 
-DownstreamAddrGroup *Http2Session::get_downstream_addr_group() const {
-  return group_.get();
+const std::shared_ptr<DownstreamAddrGroup> &
+Http2Session::get_downstream_addr_group() const {
+  return group_;
 }
 
 void Http2Session::add_to_avail_freelist() {
@@ -2204,6 +2258,9 @@ void Http2Session::on_timeout() {
     break;
   }
   case CONNECTING: {
+    SSLOG(WARN, this) << "Connect time out; addr="
+                      << util::to_numeric_addr(&addr_->addr);
+
     downstream_failure(addr_);
     break;
   }
